# Copyright 2024-2025 Sébastien Demanou. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any
from typing import Generic
from typing import TypeVar

from pydantic import BaseModel

from .document import Document
from .document import UserMeta

SettingValue = TypeVar('SettingValue')


class Setting(BaseModel, Generic[SettingValue]):
  value: SettingValue


class SettingMeta(UserMeta):
  property: str
  device: str | None = None


GenericSetting = Setting[Any]
SettingDocumentType = Document[GenericSetting, SettingMeta]
SettingDocument = SettingDocumentType.create_alias_from_base_meta(
  data_cls=Setting, meta_cls=SettingMeta
)



# Copyright 2024-2025 Sébastien Demanou. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from collections.abc import Callable
from typing import Any
from typing import Literal

from pydantic import BaseModel
from pydantic import Field

from .document import Document
from .document import UserMeta


class NotificationAction(BaseModel):
  """A class representing an action associated with a notification.

  Attributes:
    label (str): The label or text of the action.
    trigger (Callable[[], None]): The trigger function that will be called when the action is activated.
  """

  label: str
  trigger: Callable[[], None]


class Notification(BaseModel):
  """
  A class representing a notification.

  Attributes:
      summary (str): The summary of the notification.
      body (str): The body text of the notification.
      icon (str): The resolved file path to the icon for the notification.
      sound (str | None): The filename of the notification sound, or None if no sound is specified.
      actions (list[NotificationAction]): The list of actions for the notification.
      timeout (int): The timeout duration for the notification in milliseconds.
                      0: The notification never expires.
                      Positive integer: The timeout time in milliseconds since the display of the notification
                          at which the notification should automatically close.
      urgency (Literal[0, 1, 2]): The urgency level of the notification.
                      0: Low
                      1: Normal
                      2: Critical.
      on_close (Callable[[], Any] | None): The callback function to be called when the notification is closed,
                                            or None if no callback is specified.
  """

  summary: str
  body: str
  icon: str | None = None
  sound: str | None = None
  actions: list[NotificationAction] = Field(default_factory=list)
  timeout: int = 5000  # 5 seconds
  urgency: Literal[0, 1, 2] = 0
  on_close: Callable[[], Any] | None = None

  def __str__(self):
    return self.body


class Announcement(BaseModel):
  """A class representing an announcement.

  Attributes:
    icon (str | None): The icon associated with the announcement, or None if no icon is specified.
    message (str): The announcement message.
    sound (str | None): The filename of the notification sound, or None if no sound is specified.
    repeat (int): The number of times the announcement should be repeated. Default is 1.
  """

  message: str
  icon: str | None = None
  sound: str | None = None
  repeat: int = 1


class Track(BaseModel):
  media_url: str
  title: str | None = None
  artist_name: str | None = None
  album_title: str | None = None
  album_artist: str | None = None
  description: str | None = None
  artwork_url: str | None = None


class PrintTextAction(BaseModel):
  type: Literal['print'] = 'print'
  text: str


class PrintErrorAction(BaseModel):
  type: Literal['error'] = 'error'
  kind: Literal['unexpected:error', 'processing:not-understood', 'skill:disabled']
  data: dict[str, Any] | None = None


class RequestAdditionalInfoAction(BaseModel):
  type: Literal['system:request'] = 'system:request'
  text: str
  context: dict = Field(default_factory=dict)


class PlayAction(BaseModel):
  type: Literal['media:play'] = 'media:play'
  tracks: list[Track]


class PauseAction(BaseModel):
  type: Literal['media:pause'] = 'media:pause'


class ResumeAction(BaseModel):
  type: Literal['media:resume'] = 'media:resume'


class StopAction(BaseModel):
  type: Literal['media:stop'] = 'media:stop'


class MuteAction(BaseModel):
  type: Literal['media:mute'] = 'media:mute'


class UnmuteAction(BaseModel):
  type: Literal['media:unmute'] = 'media:unmute'


class SetVolumeAction(BaseModel):
  type: Literal['media:set-volume'] = 'media:set-volume'
  value: int


class NextAction(BaseModel):
  type: Literal['media:next'] = 'media:next'


class PreviousAction(BaseModel):
  type: Literal['media:previous'] = 'media:previous'


class NotifyAction(BaseModel):
  type: Literal['system:notify'] = 'system:notify'
  notification: Notification


class AnnounceAction(BaseModel):
  type: Literal['system:announcement'] = 'system:announcement'
  announcement: Announcement


class DoneAction(BaseModel):
  type: Literal['request:done'] = 'request:done'


class CancelAction(BaseModel):
  type: Literal['request:cancel'] = 'request:cancel'


UnexpectedErrorType = Literal['error']
InternalProcessingStart = Literal['processing:start']
InternalProcessingEnd = Literal['processing:end']

ActionType = Literal[
  'print',
  'error',
  'system:request',
  'media:play',
  'media:pause',
  'media:resume',
  'media:stop',
  'media:mute',
  'media:unmute',
  'media:set-volume',
  'media:next',
  'media:previous',
  'system:notify',
  'system:announcement',
  'request:done',
  'request:cancel',
]


ProcessingAction = (
  PrintTextAction
  | PrintErrorAction
  | PlayAction
  | PauseAction
  | ResumeAction
  | StopAction
  | MuteAction
  | UnmuteAction
  | SetVolumeAction
  | NextAction
  | PreviousAction
  | NotifyAction
  | AnnounceAction
  | RequestAdditionalInfoAction
  | DoneAction
  | CancelAction
)

Actions = list[ProcessingAction]


class ClassificationPrediction(BaseModel):
  """
  A class representing a prediction for a classification task.

  Attributes:
    label (str): The predicted label.
    accuracy (float): The predicted accuracy.
  """

  label: str
  accuracy: float


class Request(BaseModel):
  type: Literal['request']
  input: Literal['text', 'audio']
  processed_text: str
  prediction: ClassificationPrediction | None
  actions: Actions


DetectionType = Literal['wakeword', 'sound']


class Detection(BaseModel):
  type: DetectionType
  sound: str
  score: float


class ActivityMeta(UserMeta, BaseModel):
  """
  A class representing metadata for an activity.
  """

  device: str


DetectionDocumentType = Document[Detection, ActivityMeta]
DetectionDocument = DetectionDocumentType.create_alias_from_user_meta(
  data_cls=Detection, meta_cls=ActivityMeta
)

RequestDocumentType = Document[Request, ActivityMeta]
RequestDocument = RequestDocumentType.create_alias_from_user_meta(
  data_cls=Request, meta_cls=ActivityMeta
)



# Copyright 2024-2025 Sébastien Demanou. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import json
import uuid
from datetime import datetime
from typing import Any
from typing import Generic
from typing import TypeVar

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field


class BaseMeta(BaseModel):
  created_at: str = Field(default_factory=lambda: datetime.now().isoformat())
  updated_at: str | None = Field(default=None)


class UserMeta(BaseModel):
  uid: str
  created_at: str = Field(default_factory=lambda: datetime.now().isoformat())
  updated_at: str | None = Field(default=None)


Data = TypeVar('Data', bound=BaseModel | dict)
Meta = TypeVar('Meta', bound=BaseMeta | UserMeta)


class Document(BaseModel, Generic[Data, Meta]):
  id: str = Field(default_factory=lambda: str(uuid.uuid4()), alias='_id')
  rev: str | None = Field(alias='_rev', default=None)
  type: str
  data: Data
  meta: Meta

  model_config = ConfigDict(
    arbitrary_types_allowed=True,
    populate_by_name=True,
    extra='ignore',
  )

  @classmethod
  def create_alias_from_user_meta(
    cls: type['Document[Data, Meta]'],
    data_cls: type[Data],
    meta_cls: type[UserMeta] = UserMeta,
  ) -> 'Document[Data, UserMeta]':
    return cls.create_alias_from_base_meta(
      data_cls=data_cls,
      meta_cls=meta_cls,
    )

  @classmethod
  def create_alias_from_base_meta(
    cls: type['Document[Data, Meta]'],
    data_cls: type[Data],
    meta_cls: type[BaseMeta] = BaseMeta,
  ) -> 'Document[Data, Meta]':
    class WrappedDocument(BaseModel):
      @staticmethod
      def model_validate(obj: dict[str, Any], **kwargs):
        data = data_cls.model_validate(obj['data']) if isinstance(data_cls, dict) else obj['data']
        meta = meta_cls.model_validate(obj['meta'])

        return cls.model_validate({**obj, 'data': data, 'meta': meta}, **kwargs)

      @classmethod
      def model_validate_json(cls, json_data: str, **kwargs):
        data_dict = json.loads(json_data)
        return WrappedDocument.model_validate(data_dict, **kwargs)

      def model_dump(self, by_alias: bool = True, **kwargs) -> dict[str, Any]:
        return super().model_dump(by_alias=by_alias, **kwargs)

    return WrappedDocument


GenericDocumentType = Document[dict, BaseMeta]
GenericDocument = GenericDocumentType.create_alias_from_base_meta(data_cls=dict)

GenericUserDocumentType = Document[dict, UserMeta]
GenericUserDocument = GenericUserDocumentType.create_alias_from_user_meta(data_cls=dict)



# Copyright 2024-2025 Sébastien Demanou. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any
from typing import Generic
from typing import TypeVar

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field

DocumentViewKey = TypeVar('DocumentViewKey')
DocumentViewValue = TypeVar('DocumentViewValue')
DocumentViewDoc = TypeVar('DocumentViewDoc', bound=dict | BaseModel)


class DocumentViewRow(
  BaseModel,
  Generic[DocumentViewKey, DocumentViewValue, DocumentViewDoc],
):
  id: str
  key: DocumentViewKey
  value: DocumentViewValue
  doc: DocumentViewDoc | None = None


class DocumentViewRowError(BaseModel, Generic[DocumentViewKey]):
  key: DocumentViewKey
  error: str


class DocumentViewResponse(
  BaseModel,
  Generic[DocumentViewKey, DocumentViewValue, DocumentViewDoc],
):
  """
  View response.

  Docs: http://docs.couchdb.org/en/latest/api/ddoc/views.html#get--db-_design-ddoc-_view-view
  """

  offset: int | None
  total_rows: int
  rows: list[DocumentViewRow[DocumentViewKey, DocumentViewValue, DocumentViewDoc]]
  update_seq: Any | None = None


class GenericDocumentViewResponse(BaseModel):
  """
  View response.

  Docs: http://docs.couchdb.org/en/latest/api/ddoc/views.html#get--db-_design-ddoc-_view-view
  """

  offset: int | None
  total_rows: int
  rows: list[dict[str, Any]]
  update_seq: Any | None = None


class FindResponse(BaseModel):
  docs: list[dict]
  warning: str | None = None
  execution_stats: dict | None = None
  bookmark: str | None = None


class GenericResponse(BaseModel):
  ok: bool
  id: str
  rev: str


class OkResponse(BaseModel):
  ok: bool


class GetNodeVersionResponse(BaseModel):
  pass


class MaybeDocument(BaseModel):
  id: str = Field(alias='_id')
  rev: str | None = Field(alias='_rev', default=None)

  model_config = ConfigDict(
    arbitrary_types_allowed=True,
    populate_by_name=True,
    extra='forbid',
  )


class IdentifiedDocument(BaseModel):
  id: str = Field(alias='_id')

  model_config = ConfigDict(
    arbitrary_types_allowed=True,
    populate_by_name=True,
    extra='forbid',
  )


class RevisionedDocument(BaseModel):
  rev: str = Field(alias='_rev')

  model_config = ConfigDict(
    arbitrary_types_allowed=True,
    populate_by_name=True,
    extra='forbid',
  )


class Document(IdentifiedDocument, RevisionedDocument, BaseModel):
  pass


class FilterDocument(IdentifiedDocument, BaseModel):
  filters: dict[str, str]
  language: str | None = None


DesignDocumentView = dict[str, str]


class DesignDocument(IdentifiedDocument, BaseModel):
  views: dict[str, DesignDocumentView]
  language: str | None = None


class CreateSessionResponse(BaseModel):
  ok: bool
  name: str
  roles: list[str]


class SessionInfo(BaseModel):
  authenticated: str
  authentication_handlers: list[str]
  authentication_db: str | None = None


class UserCtx(BaseModel):
  name: str
  roles: list[str]


class GetSessionResponse(BaseModel):
  info: SessionInfo
  ok: bool
  userCtx: UserCtx
